/* 
  BMP180 Pressure Sensor Wrapper for Teensy 3.x
  -------------------------------------------------------------
  
  Written 2016 By Bernd Wokoeck
  
  Version history:
  1.00   02/21/2016  created

  Dependencies:
  - Wire-Library

  Derived from Adafruit_BMP085_U.h/Adafruit_BMP085_U.cpp
  Copyright notice:
  --------------------------------------------------------------------------
  Designed specifically to work with the Adafruit BMP085 or BMP180 Breakout 
  ----> http://www.adafruit.com/products/391
  ----> http://www.adafruit.com/products/1603
 
  These displays use I2C to communicate, 2 pins are required to interface.

  Adafruit invests time and resources providing this open source code,
  please support Adafruit andopen-source hardware by purchasing products
  from Adafruit!

  Written by Kevin Townsend for Adafruit Industries.  
  BSD license, all text above must be included in any redistribution
  --------------------------------------------------------------------------
**************************************************************/

#include "Bmp180Sensor.h"

void Bmp180Sensor::Init( uint32_t period )
{
  m_period = max( 30u, period - 20u );
  m_temp = 0.0f;
  m_pressure = 0.0f;
  m_rawTemp = 0;
  m_currentJob.Init(); 

  // Wire.setDefaultTimeout( 500 );

  // Enable I2C
  Wire.begin();

  // Make sure we have the right device
  uint8_t id;
  read8( BMP085_REGISTER_CHIPID, &id );
  if(id != 0x55)
    return;

  // Coefficients need to be read once
  readCoefficients();

  m_bInitalized = true;
}

void Bmp180Sensor::DoBmp180Sensor()
{
  if( !m_bInitalized )
    return;

  // START
  if( m_currentJob.IsIdle() )
  {
    if( !m_currentJob.IsPressure() )
    {
      // start pressure measurement
      m_currentJob.Start( Bmp180Sensor::Job::PRESSURE, 26 );   // wait 26 ms for sensor
      writeCommand(BMP085_REGISTER_CONTROL, BMP085_REGISTER_READPRESSURECMD + (BMP085_MODE_ULTRAHIGHRES << 6));
    }
    else
    {
      m_currentJob.Start( Bmp180Sensor::Job::TEMP, 5 );    // wait 5 ms for sensor
      writeCommand(BMP085_REGISTER_CONTROL, BMP085_REGISTER_READTEMPCMD);
    }
  }
  // WAITING
  else if ( m_currentJob.IsWaiting() )
  {
    // wait until measure time is past due
    if( millis() > m_currentJob.m_endTime )
    {
      if( m_currentJob.IsPressure() )
      {
        // read pressure result
        uint8_t  p8;
        uint16_t p16;
        int32_t  p32;
        read16(BMP085_REGISTER_PRESSUREDATA, &p16);
        p32 = (uint32_t)p16 << 8;
        read8(BMP085_REGISTER_PRESSUREDATA+2, &p8);
        p32 += p8;
        p32 >>= (8 - BMP085_MODE_ULTRAHIGHRES);
        m_pressure = raw2Pressure( p32, m_rawTemp );
      }
      else
      {
        // read temp result
        uint16_t t;
        read16(BMP085_REGISTER_TEMPDATA, &t);
        m_rawTemp = t;
        m_temp    = raw2Temperature( t );
      }
      m_currentJob.SetSleeping( m_period );
    }
  }
  // SLEEP
  else if ( m_currentJob.IsSleeping() )
  {
    if( millis() > m_currentJob.m_endTime )
      m_currentJob.SetIdle();
  }
}


long  Bmp180Sensor::GetAlt( float seaLevelPressure )
{
  if( m_pressure < 100.0f || m_pressure > 1100.0f )
    return -1;

  long alt = pressure2Alt(seaLevelPressure, m_pressure );

  if( m_altGroundLevel < 0 )
    m_altGroundLevel = alt;

  alt -= m_altGroundLevel;
  
  if( alt == -1 ) // -1 is "invalid value", avoid it
    alt = 0;
  
  return alt;
}


/**************************************************************************/
/*!
    @brief  Gets the compensated pressure level in kPa
*/
/**************************************************************************/
float Bmp180Sensor::raw2Pressure( uint32_t rawPressure, uint32_t rawTemp )
{
  int32_t  ut = 0, up = 0, compp = 0;
  int32_t  x1, x2, b5, b6, x3, b3, p;
  uint32_t b4, b7;

  /* Get the raw pressure and temperature values */
  ut = rawTemp;
  up = rawPressure;

  /* Temperature compensation */
  b5 = computeB5(ut);

  /* Pressure compensation */
  b6 = b5 - 4000;
  x1 = (m_bmp180_coeffs.b2 * ((b6 * b6) >> 12)) >> 11;
  x2 = (m_bmp180_coeffs.ac2 * b6) >> 11;
  x3 = x1 + x2;
  b3 = (((((int32_t) m_bmp180_coeffs.ac1) * 4 + x3) << BMP085_MODE_ULTRAHIGHRES ) + 2) >> 2;
  x1 = (m_bmp180_coeffs.ac3 * b6) >> 13;
  x2 = (m_bmp180_coeffs.b1 * ((b6 * b6) >> 12)) >> 16;
  x3 = ((x1 + x2) + 2) >> 2;
  b4 = (m_bmp180_coeffs.ac4 * (uint32_t) (x3 + 32768)) >> 15;
  b7 = ((uint32_t) (up - b3) * (50000 >> BMP085_MODE_ULTRAHIGHRES));

  if (b7 < 0x80000000)
  {
    p = (b7 << 1) / b4;
  }
  else
  {
    p = (b7 / b4) << 1;
  }

  x1 = (p >> 8) * (p >> 8);
  x1 = (x1 * 3038) >> 16;
  x2 = (-7357 * p) >> 16;
  compp = p + ((x1 + x2 + 3791) >> 4);

  /* Assign compensated pressure value */
  return (float)compp/100.0F;
}

/**************************************************************************/
/*!
    @brief  Reads the temperatures in degrees Celsius
*/
/**************************************************************************/
float Bmp180Sensor::raw2Temperature(  uint32_t rawTemp  )
{
  int32_t B5;
  float t;

  B5 = computeB5( rawTemp );
  t = (B5+8) >> 4;
  t /= 10;

  return t;
}

/**************************************************************************/
/*!
    Calculates the altitude (in meters) from the specified atmospheric
    pressure (in hPa), and sea-level pressure (in hPa).

    @param  seaLevel      Sea-level pressure in hPa
    @param  atmospheric   Atmospheric pressure in hPa
*/
/**************************************************************************/
float Bmp180Sensor::pressure2Alt(float seaLevel, float atmospheric)
{
  // Equation taken from BMP180 datasheet (page 16):
  //  http://www.adafruit.com/datasheets/BST-BMP180-DS000-09.pdf

  // Note that using the equation from wikipedia can give bad results
  // at high altitude.  See this thread for more information:
  //  http://forums.adafruit.com/viewtopic.php?f=22&t=58064
  
  return 44330.0 * (1.0 - pow(atmospheric / seaLevel, 0.1903f));
}

/**************************************************************************/
/*!
    Calculates the altitude (in meters) from the specified atmospheric
    pressure (in hPa), and sea-level pressure (in hPa).  Note that this
    function just calls the overload of pressureToAltitude which takes
    seaLevel and atmospheric pressure--temperature is ignored.  The original
    implementation of this function was based on calculations from Wikipedia
    which are not accurate at higher altitudes.  To keep compatibility with
    old code this function remains with the same interface, but it calls the
    more accurate calculation.

    @param  seaLevel      Sea-level pressure in hPa
    @param  atmospheric   Atmospheric pressure in hPa
    @param  temp          Temperature in degrees Celsius
*/
/**************************************************************************/
float Bmp180Sensor::pressure2Alt(float seaLevel, float atmospheric, float temp)
{
  return pressure2Alt(seaLevel, atmospheric);
}

/**************************************************************************/
/*!
    Calculates the pressure at sea level (in hPa) from the specified altitude 
    (in meters), and atmospheric pressure (in hPa).  

    @param  altitude      Altitude in meters
    @param  atmospheric   Atmospheric pressure in hPa
*/
/**************************************************************************/
float Bmp180Sensor::seaLevelForAlt(float altitude, float atmospheric)
{
  // Equation taken from BMP180 datasheet (page 17):
  //  http://www.adafruit.com/datasheets/BST-BMP180-DS000-09.pdf

  // Note that using the equation from wikipedia can give bad results
  // at high altitude.  See this thread for more information:
  //  http://forums.adafruit.com/viewtopic.php?f=22&t=58064
  
  return atmospheric / pow(1.0 - (altitude/44330.0), 5.255);
}

/**************************************************************************/
/*!
    Calculates the pressure at sea level (in hPa) from the specified altitude 
    (in meters), and atmospheric pressure (in hPa).  Note that this
    function just calls the overload of seaLevelForAltitude which takes
    altitude and atmospheric pressure--temperature is ignored.  The original
    implementation of this function was based on calculations from Wikipedia
    which are not accurate at higher altitudes.  To keep compatibility with
    old code this function remains with the same interface, but it calls the
    more accurate calculation.

    @param  altitude      Altitude in meters
    @param  atmospheric   Atmospheric pressure in hPa
    @param  temp          Temperature in degrees Celsius
*/
/**************************************************************************/
float Bmp180Sensor::seaLevelForAlt(float altitude, float atmospheric, float temp)
{
  return seaLevelForAlt(altitude, atmospheric);
}

/**************************************************************************/
/*!
    @brief  Compute B5 coefficient used in temperature & pressure calcs.
*/
/**************************************************************************/
int32_t Bmp180Sensor::computeB5(int32_t ut)
{
  int32_t X1 = (ut - (int32_t)m_bmp180_coeffs.ac6) * ((int32_t)m_bmp180_coeffs.ac5) >> 15;
  int32_t X2 = ((int32_t)m_bmp180_coeffs.mc << 11) / (X1+(int32_t)m_bmp180_coeffs.md);
  return X1 + X2;
}

/**************************************************************************/
/*!
    @brief  Writes an 8 bit value over I2C
*/
/**************************************************************************/
void Bmp180Sensor::writeCommand(byte reg, byte value)
{
  Wire.beginTransmission((uint8_t)BMP180_ADDRESS);
  Wire.write((uint8_t)reg);
  Wire.write((uint8_t)value);
  Wire.endTransmission();
}

/**************************************************************************/
/*!
    @brief  Reads an 8 bit value over I2C
*/
/**************************************************************************/
void Bmp180Sensor::read8(byte reg, uint8_t *value)
{
  Wire.beginTransmission((uint8_t)BMP180_ADDRESS);
  Wire.write((uint8_t)reg);
  Wire.endTransmission();
  Wire.requestFrom((uint8_t)BMP180_ADDRESS, (byte)1);
  *value = Wire.read();
  Wire.endTransmission();
}

/**************************************************************************/
/*!
    @brief  Reads a 16 bit value over I2C
*/
/**************************************************************************/
void Bmp180Sensor::read16(byte reg, uint16_t *value)
{
  Wire.beginTransmission((uint8_t)BMP180_ADDRESS);
  Wire.write((uint8_t)reg);
  Wire.endTransmission();
  Wire.requestFrom((uint8_t)BMP180_ADDRESS, (byte)2);
  *value = (Wire.read() << 8) | Wire.read();
  Wire.endTransmission();
}

/**************************************************************************/
/*!
    @brief  Reads a signed 16 bit value over I2C
*/
/**************************************************************************/
void Bmp180Sensor::readS16(byte reg, int16_t *value)
{
  uint16_t i;
  read16(reg, &i);
  *value = (int16_t)i;
}

/**************************************************************************/
/*!
    @brief  Reads the factory-set coefficients
*/
/**************************************************************************/
void Bmp180Sensor::readCoefficients(void)
{
  readS16(BMP085_REGISTER_CAL_AC1, &m_bmp180_coeffs.ac1);
  readS16(BMP085_REGISTER_CAL_AC2, &m_bmp180_coeffs.ac2);
  readS16(BMP085_REGISTER_CAL_AC3, &m_bmp180_coeffs.ac3);
  read16(BMP085_REGISTER_CAL_AC4, &m_bmp180_coeffs.ac4);
  read16(BMP085_REGISTER_CAL_AC5, &m_bmp180_coeffs.ac5);
  read16(BMP085_REGISTER_CAL_AC6, &m_bmp180_coeffs.ac6);
  readS16(BMP085_REGISTER_CAL_B1, &m_bmp180_coeffs.b1);
  readS16(BMP085_REGISTER_CAL_B2, &m_bmp180_coeffs.b2);
  readS16(BMP085_REGISTER_CAL_MB, &m_bmp180_coeffs.mb);
  readS16(BMP085_REGISTER_CAL_MC, &m_bmp180_coeffs.mc);
  readS16(BMP085_REGISTER_CAL_MD, &m_bmp180_coeffs.md);
}
