/* 
  BMP180 Pressure Sensor Wrapper for Teensy 3.x
  -------------------------------------------------------------
  
  Copyright (C) 2015 Bernd Wokoeck
  
  Version history:
  1.00   02/21/2016  created

  Dependencies:
  - Wire-Library
  
**************************************************************/
#if ARDUINO >= 100
 #include "Arduino.h"
#else
 #include "WProgram.h"
#endif

#include <Wire.h> 
// #include <i2c_t3.h> // see https://forum.pjrc.com/threads/21680-New-I2C-library-for-Teensy3

class Bmp180Sensor
{
public:
  Bmp180Sensor()  { m_bInitalized = false; m_altGroundLevel = -1; }
  ~Bmp180Sensor() {}

  void Init( uint32_t period = 100 ); // default period 100ms = 10 Hz 
  void DoBmp180Sensor();

  long  GetAlt( float seaLevelPressure = 1013.0f );
  float GetTemp()     { return m_temp; }
  float GetPressure() { return m_pressure; }

  // conversions
  float pressure2Alt(float seaLevel, float atmospheric);
  float pressure2Alt(float seaLevel, float atmospheric, float temp);
  float seaLevelForAlt(float altitude, float atmospheric);
  float seaLevelForAlt(float altitude, float atmospheric, float temp);

protected:
  enum
  {
    BMP180_ADDRESS = 0x77,
  };
    

  // MODE SETTINGS
  enum
  {
    BMP085_MODE_ULTRALOWPOWER          = 0,
    BMP085_MODE_STANDARD               = 1,
    BMP085_MODE_HIGHRES                = 2,
    BMP085_MODE_ULTRAHIGHRES           = 3
  };

  // REGISTERS
  enum
  {
    BMP085_REGISTER_CAL_AC1            = 0xAA,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_AC2            = 0xAC,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_AC3            = 0xAE,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_AC4            = 0xB0,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_AC5            = 0xB2,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_AC6            = 0xB4,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_B1             = 0xB6,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_B2             = 0xB8,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_MB             = 0xBA,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_MC             = 0xBC,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CAL_MD             = 0xBE,  // R   Calibration data (16 bits)
    BMP085_REGISTER_CHIPID             = 0xD0,
    BMP085_REGISTER_VERSION            = 0xD1,
    BMP085_REGISTER_SOFTRESET          = 0xE0,
    BMP085_REGISTER_CONTROL            = 0xF4,
    BMP085_REGISTER_TEMPDATA           = 0xF6,
    BMP085_REGISTER_PRESSUREDATA       = 0xF6,
    BMP085_REGISTER_READTEMPCMD        = 0x2E,
    BMP085_REGISTER_READPRESSURECMD    = 0x34
  };

  // CALIBRATION DATA
  typedef struct
  {
    int16_t  ac1;
    int16_t  ac2;
    int16_t  ac3;
    uint16_t ac4;
    uint16_t ac5;
    uint16_t ac6;
    int16_t  b1;
    int16_t  b2;
    int16_t  mb;
    int16_t  mc;
    int16_t  md;
  }
  bmp180_calib_data;

  bmp180_calib_data m_bmp180_coeffs;

  // conversions
  float raw2Temperature( uint32_t rawTemp );
  float raw2Pressure( uint32_t rawPressure, uint32_t rawTemp );
  int32_t computeB5(int32_t ut);

  // low level I2C commands
  void writeCommand(byte reg, byte value);
  void read8(byte reg, uint8_t *value);
  void read16(byte reg, uint16_t *value);
  void readS16(byte reg, int16_t *value);
  void readCoefficients(void);

  // job data
  class Job
  {
    friend class Bmp180Sensor;
  public:
    Job() {}

  protected:
    void Init() { m_endTime = 0; m_measureType = PRESSURE; m_status = IDLE; }

    enum enStatus
    {
      IDLE     = 0, // measurement can be started 
      WAITING  = 1, // wait for result
      SLEEPING = 2, // wait until next measurement, determines rate
    };
    enum enMeasureType
    {
      PRESSURE = 0,
      TEMP     = 1,
    };

    enStatus      m_status;
    enMeasureType m_measureType;
    uint32_t      m_endTime;

  public:
    void  Start( enMeasureType enType, uint32_t waitMillis )
    {
      m_measureType = enType;
      m_status =  WAITING;
      m_endTime = millis() + waitMillis;
    }
    void SetSleeping( uint32_t waitMillis )
    {
      m_status =  SLEEPING;
      m_endTime = millis() + waitMillis;
    }
    inline void SetIdle()  { m_status =  IDLE; }

    inline bool     IsIdle()     { return m_status == IDLE; }
    inline bool     IsWaiting()  { return m_status == WAITING; }
    inline bool     IsSleeping() { return m_status == SLEEPING; }
    inline bool     IsPressure() { return m_measureType == PRESSURE; }
  };

  bool     m_bInitalized;
  Job      m_currentJob;
  uint32_t m_rawTemp;
  uint32_t m_period;  // measurement rate

  float    m_temp;
  float    m_pressure;

  long    m_altGroundLevel;
};
