/* 
  Voltage Sensor Wrapper for Teensy 3.x
  -------------------------------------------------------------
  
  Copyright 2016 By Bernd Wokoeck
  
  Version history:
  1.00   02/21/2016  created

  Dependencies:
  - ADC library for Teensy https://github.com/pedvide/ADC

**************************************************************/

#include "DualVoltageSensor.h"

#ifdef CORE_TEENSY
  #include <ADC.h>
  ADC *p_adc = new ADC(); // adc object;
#endif 


void DualVoltageSensor::Init( int pin1, int pin2, float conv1, float conv2, int period  )
{
  pinMode( pin1, INPUT );
  pinMode( pin2, INPUT );
  
  m_period = period;
  m_pin1   = pin1;
  m_pin2   = pin2;

  m_convFactor1 = conv1;
  m_convFactor2 = conv2;

#ifdef CORE_TEENSY
  // reference can be ADC_REF_3V3, ADC_REF_1V2 (not for Teensy LC) or ADC_REF_EXT.
  p_adc->setReference( ADC_REF_3V3, ADC_0 ); // change all 3.3 to 1.2 if you change the reference to 1V2
  p_adc->setAveraging( 32, ADC_0 ); // set number of averages
  p_adc->setResolution( 12, ADC_0 ); // set bits of resolution
  p_adc->setConversionSpeed( ADC_VERY_LOW_SPEED, ADC_0 ); // change the conversion speed
  p_adc->setSamplingSpeed( ADC_VERY_LOW_SPEED, ADC_0 ); // change the sampling speed

  p_adc->setReference( ADC_REF_3V3, ADC_1 ); // change all 3.3 to 1.2 if you change the reference to 1V2
  p_adc->setAveraging( 32, ADC_1 ); // set number of averages
  p_adc->setResolution (12, ADC_1 ); // set bits of resolution
  p_adc->setConversionSpeed( ADC_VERY_LOW_SPEED, ADC_1 ); // change the conversion speed
  p_adc->setSamplingSpeed( ADC_VERY_LOW_SPEED, ADC_1 ); // change the sampling speed
#else
  // Select Vref=AVcc
  ADMUX |= (1<<REFS0);
  //set prescaller to 128 and enable ADC 
  ADCSRA |= (1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)|(1<<ADEN);    
#endif

}


void DualVoltageSensor::DoDualVoltageSensor()
{
  if( m_status == IDLE )
  {
    StartConversion( m_channel );
    m_status = CONVERTING;
  }
  else if( m_status == CONVERTING )
  {
    if( IsReady() )
    {
      m_val = (long)(((float)ReadResult()) * m_factor);
      m_status = SLEEPING;
      m_endTime = millis() + m_period;
      m_channel ^= 1;
    }
  }
  else if( m_status == SLEEPING )
  {
    if( millis() > m_endTime )
      m_status = IDLE;
  }
}


void DualVoltageSensor::StartConversion( uint8_t mux  )
{
#ifdef CORE_TEENSY
  if( m_channel == 0 )
  {
    p_adc->startSingleRead( m_pin1 );
    m_factor = m_convFactor1; 
  }
  else
  {
    p_adc->startSingleRead( m_pin2 );
    m_factor = m_convFactor2; 
  }
#else
 //select ADC channel with safety mask
 ADMUX = (ADMUX & 0xF0) | (mux & 0x0F);
 //single conversion mode
 ADCSRA |= (1<<ADSC);
#endif
}


uint32_t DualVoltageSensor::ReadResult()
{
#ifdef CORE_TEENSY
  return p_adc->readSingle();
#else
  return ADC;
#endif
}


bool DualVoltageSensor::IsReady()
{
#ifdef CORE_TEENSY
  return p_adc->isComplete();
#else
  return (ADCSRA & (1<<ADSC)) ? true : false;
#endif
}

