/* 
  Jeti EX MegaSensor for Teensy 3.x
  -------------------------------------------------------------
  
  Copyright (C) 2017 Bernd Wokoeck
  
  Version history:
  1.00   02/21/2016  created
  1.50   04/08/2017  bidirectional terminal communication added
                     ECU telemetry values added:
                     EGT average, EGT max, STR spin down time, Total runtime
  1.51   05/01/2017  quick and dirty fix for sporadic missing total runtime
                     in OrbitECUSensor

  Dependencies:
  - OrbitECU telemetry library for AECU Version T17
  - TinyGPS
  - JetiExSensor-library
    see https://sourceforge.net/projects/jetiexsensorcpplib/
  
**************************************************************/

#include <TinyGPS.h>
#include <OrbitECUSensor.h>
#include <EEPROM.h>
#include <Wire.h> 
#include <JetiExSerial.h>
#include <JetiExProtocol.h>
#include <ADC.h>

#include "GpsSensor.h"
#include "Bmp180Sensor.h"
#include "DualVoltageSensor.h"

DualVoltageSensor  voltage;
Bmp180Sensor       pressure;
GpsSensor          gps;
OrbitECUSensor     ecu;
JetiExProtocol     jetiEx;

enum
{
  ID_VOLTAGE  = 1,  // ECU voltage
  ID_EGT      = 2,  // ECU turbine exhaust gas temperature (EGT)
  ID_RPM      = 3,  // ECU turbine RPM x 1000
  ID_FUEL     = 4,  // ECU consumed fuel in ml, calculated by fuel pump voltage
  ID_STATUS   = 5,  // ECU status value

  ID_GPSLAT   = 6,  // GPS latitude xx.xxxxx degrees
  ID_GPSLON   = 7,  // GPS longitude xx.xxxxx degrees
  ID_GPSSPEED = 8,  // GPS speed in km/h
  ID_GPSHDOP  = 9,  // GPS horizonal dilution of precision 0.00...99.99
  ID_ALT      = 10, // altitude from GPS or pressure sensor

  ID_BATT     = 11, // battery voltage batt1/batt2 of RC system
  
  ID_PUMP     = 12, // pump voltage
  ID_EGTMAX   = 13, // maximum EGT
  ID_TEMPAMB  = 14, // ambient temperature (pressure sensor)

  ID_EAT      = 15, // exhaust gas average temp (ECU) 
  ID_EMT      = 16, // exhaust gas max temp (ECU) 
  ID_TISPINDN = 17, // spin down time in 0.1s
  ID_TOTTIME  = 18, // total runtime h:m
  // ID_GASPERC  = 19, // gas percentage
};

// id from 1..15
// name plus unit must be < 20 characters
// precision = 0 --> 0, precision = 1 --> 0.0, precision = 2 --> 0.00
JETISENSOR_CONST sensors[] PROGMEM = 
{
  //              id             name        unit         data type             precision 
 { ID_VOLTAGE,    "ECU Batt.",  "V",         JetiSensor::TYPE_14b, 1 },
 { ID_EGT,        "EGT",        "\xB0\x43",  JetiSensor::TYPE_14b, 0 }, // �C
 { ID_RPM,        "RPM x 1000", "/min",      JetiSensor::TYPE_14b, 1 },
 { ID_FUEL,       "Fuel",       "ml",        JetiSensor::TYPE_14b, 0 },
 { ID_STATUS,     "Status",     " ",         JetiSensor::TYPE_14b, 0 },

 { ID_GPSLAT,     "Latitude",   " ",         JetiSensor::TYPE_GPS, 0 },
 { ID_GPSLON,     "Longitude",  " ",         JetiSensor::TYPE_GPS, 0 },
 { ID_GPSSPEED,   "Speed",      "km/h",      JetiSensor::TYPE_14b, 0 },
 { ID_GPSHDOP,    "Gps HDOP",   " ",         JetiSensor::TYPE_14b, 2 },
 { ID_ALT,        "Altitude",   "m",         JetiSensor::TYPE_14b, 0 },

 { ID_BATT,       "RC Batt.",   "V",         JetiSensor::TYPE_14b, 2 },

 { ID_PUMP,       "Pump",       "V",         JetiSensor::TYPE_14b, 2 },
 { ID_EGTMAX,     "EGTmax",     "\xB0\x43",  JetiSensor::TYPE_14b, 0 }, // �C
 { ID_TEMPAMB,    "Temp",       "\xB0\x43",  JetiSensor::TYPE_14b, 1 }, // �C 

 { ID_EAT,        "EGT Avg",    "\xB0\x43",  JetiSensor::TYPE_14b, 0 }, // �C
 { ID_EMT,        "EGT Max",    "\xB0\x43",  JetiSensor::TYPE_14b, 0 }, // �C
 { ID_TISPINDN,   "Spin down",  "s",         JetiSensor::TYPE_14b, 1 },
 { ID_TOTTIME,    "Total time", "H:M",       JetiSensor::TYPE_DT,  0 },
 // { ID_GASPERC,    "Gas",        "%s",        JetiSensor::TYPE_14b, 0 },
 { 0 }// end of array
};

void setup()
{
  Serial.begin( 9600 );

  voltage.Init( A0, A1, 0.295389564f, 0.294547398f, 2000 );    // 2 batteries on A0 and A1, 2 seconds display change interval
  pressure.Init( 1000 );                                       // I2C
  ecu.Init( OrbitECUSensor::SERIAL1, false );
  gps.Init( GpsSensor::SERIAL3 );

  jetiEx.SetDeviceId( 0x76, 0x32 );     // 0x3276
  jetiEx.Start( "Lama", sensors, JetiExProtocol::SERIAL2 );
}

void loop()
{
  ecu.DoOrbitECUSensor();
  voltage.DoDualVoltageSensor(); 
  pressure.DoBmp180Sensor(); 
  gps.DoGpsSensor(); 

  int32_t status = ecu.GetStatus();
  int32_t rpm    = ecu.GetRPM();
  jetiEx.SetSensorValue( ID_VOLTAGE,  ecu.GetVoltage() );
  jetiEx.SetSensorValue( ID_EGT,      ecu.GetEGT() ); 
  jetiEx.SetSensorValue( ID_RPM,      rpm );
  jetiEx.SetSensorValue( ID_FUEL,     ecu.GetFuel() );
  jetiEx.SetSensorValue( ID_STATUS,   status );

  float lat = gps.GetLat();
  float lon = gps.GetLon();
  jetiEx.SetSensorValueGPS( ID_GPSLAT, false, lat );
  jetiEx.SetSensorValueGPS( ID_GPSLON, true, lon ); 
  jetiEx.SetSensorValue( ID_GPSSPEED, gps.GetSpeed() );  // km/h
  jetiEx.SetSensorValue( ID_GPSHDOP,  gps.GetHDOP() );   // in 100th
  // jetiEx.SetSensorValue( ID_ALT,      gps.GetAlt() );    // m
  jetiEx.SetSensorValue( ID_ALT,      pressure.GetAlt() );    // m

  jetiEx.SetSensorValue( ID_BATT,     voltage.GetDualVoltage() );
  jetiEx.SetSensorValue( ID_PUMP,     ecu.GetPumpVolt() );
  jetiEx.SetSensorValue( ID_EGTMAX,   ecu.GetEGTmax() );
  jetiEx.SetSensorValue( ID_TEMPAMB,  (long)(pressure.GetTemp()*10.0f) );

  int32_t runMins = ecu.GetTotalMins();
  if( runMins >= 0 )
    jetiEx.SetSensorValueTime( ID_TOTTIME, 0, runMins / 60, runMins % 60 );  // shift hours--> mins, mins --> secs to get 59 hours displayed
  else
    jetiEx.SetSensorValue( ID_TOTTIME, -1L );
  
  jetiEx.SetSensorValue( ID_TISPINDN, ecu.GetSpinDownTime() );
  jetiEx.SetSensorValue( ID_EAT,      ecu.GetEAT() );
  jetiEx.SetSensorValue( ID_EMT,      ecu.GetEMT() );
  // jetiEx.SetSensorValue( ID_GASPERC,  ecu.GetGasPercent() );

  HandleJetiboxDisplay();
  HandleAutoECUTerminalKeys( status, rpm ); // press terminal keys of ECU to retrieve some turbine data
    
  /* you can do other stuff here */

  jetiEx.DoJetiSend(); 
}


void HandleJetiboxDisplay()
{
  OrbitECUSensor::FlowCalibration * pConfig = ecu.GetFlowConfig();
  static int _flow1 = 0;
  static int _flow2 = 0;

  static int  _nMenu = 0; // menu tree
  // 0 = <default display> 
  // 1 = Flow1 at 1.3 Volt
  //      << 42 ml >>
  // 2 = Flow2 at 2.9 Volt
  //      << 170 ml >>

  static char _lineBuf[16];
  static bool _bSetDisplay = false;
  static bool _bConfigChanged = false;
  static OrbitECUSensor::enKeyCode _enOrbitKey = OrbitECUSensor::TRM_KEYNONE;

  uint8_t c = jetiEx.GetJetiboxKey();

  // initialize local settings
  if( _flow1 == 0 )
  {
    _flow1 = pConfig->flow1;
    _flow2 = pConfig->flow2;
  }

  // down
  if( c == JetiExProtocol::DOWN )
  {
    if( _nMenu < 2 )
      _nMenu++;
    _bSetDisplay = true;
  }

  // up
  if( c == JetiExProtocol::UP )
  { 
    if( _nMenu > 0 )
    {
      _nMenu--;
      _bSetDisplay = true;
    }
    else
    {
       _flow2 = _flow1 = 0;
      jetiEx.SetJetiboxExit(); // todo does not exit menu
      return;
    }
  }

  // left 
  if( c == JetiExProtocol::LEFT )
  { 
    if( _nMenu == 0 )
    {
      _enOrbitKey = OrbitECUSensor::TRM_KEYDN;
      _bSetDisplay = true;
    }
    else if( _nMenu == 1 && _flow1 > 1 )
    {
      _flow1--;
      _bConfigChanged = true;
      _bSetDisplay = true;
    }
    else if( _nMenu == 2 && _flow2 > 1 ) 
    {
      _flow2--;
      _bConfigChanged = true;
      _bSetDisplay = true;
    }
  }

  // right
  if( c == JetiExProtocol::RIGHT )
  { 
    if( _nMenu == 0 )
    {
      _enOrbitKey = OrbitECUSensor::TRM_KEYUP;
      _bSetDisplay = true;
    }
    else if( _nMenu == 1 && _flow1 < 1000 )
    {
      _flow1++;
      _bConfigChanged = true;
      _bSetDisplay = true;
    }
    else if( _nMenu == 2 && _flow2 < 10000 ) 
    {
      _flow2++;
      _bConfigChanged = true;
      _bSetDisplay = true;
    }
  }

  if( _bSetDisplay )
  {
    switch( _nMenu )
    {
    case 0: 
      // orbit terminal key
      if( _enOrbitKey != OrbitECUSensor::TRM_KEYNONE )
      {
        ecu.SendKey( _enOrbitKey );
        _enOrbitKey = OrbitECUSensor::TRM_KEYNONE;
      }
      // write changed config back to EEPROM
      if( _bConfigChanged ) // todo: Diese Stelle kommt nicht dran, wenn man das Menü nicht über "Up" verlässt
      {
        _bConfigChanged = false;
        pConfig->flow1 = (float)_flow1;
        pConfig->flow2 = (float)_flow2;
        _flow2 = _flow1 = 0;
        ecu.UpdateConfig(); // write to EEPROM
      }
      break;
    case 1: 
      jetiEx.SetJetiboxText( JetiExProtocol::LINE1, "Flow1 at 1.3V  " );
      sprintf( _lineBuf, "\x7f %d ml \x7e", _flow1 );
      jetiEx.SetJetiboxText( JetiExProtocol::LINE2, _lineBuf );
      break;
    case 2: 
      jetiEx.SetJetiboxText( JetiExProtocol::LINE1, "Flow2 at 2.9V  " );
      sprintf( _lineBuf, "\x7f %d ml \x7e", _flow2 );
      jetiEx.SetJetiboxText( JetiExProtocol::LINE2, _lineBuf );
      break;
    }
    _bSetDisplay = false;
  }

  // default display shows ECU data
  if( _nMenu == 0 )
  {
    jetiEx.SetJetiboxText( JetiExProtocol::LINE1, ecu.GetLine( OrbitECUSensor::LINE1 ) );
    jetiEx.SetJetiboxText( JetiExProtocol::LINE2, ecu.GetLine( OrbitECUSensor::LINE2 ) );
  }

  return;
}


void HandleAutoECUTerminalKeys( int32_t status, int32_t rpm )
{
  enum enRunStatus
  {
    Start = 0,
    KeyDownIssued,
    KeyUpIssued,
    IsRunning,
    IsStopping,
    IsCooling,
    WaitZeroRPM,
  };

  static enRunStatus _enStatus = Start;
  static uint32_t    _tiStart  = 0;

  if( _enStatus == Start && ( status == 7 || status == 9 ) /* stopped or locked */ )
  {
    // ECU switched on
    ecu.SendKey( OrbitECUSensor::TRM_KEYDN ); 
    _enStatus = KeyDownIssued;
    _tiStart = millis();
  }
  else if( _enStatus == KeyDownIssued && millis() > _tiStart + 1000 )
  {
    // 1 second after key down
    ecu.SendKey( OrbitECUSensor::TRM_KEYUP ); 
    _enStatus = KeyUpIssued;
  }
  else if( _enStatus == KeyUpIssued &&  status == 2 /* idle */ )
  {
    _enStatus = IsRunning;
  }
  else if( _enStatus == IsRunning &&  status == 0 /* -off */ )
  {
    _enStatus = IsStopping;
  }
  else if( _enStatus == IsStopping &&  status == 8 /* cool */ )
  {
    _enStatus = IsCooling;
  }
  else if( _enStatus == IsCooling &&  status == 7 /* stop */ )
  {
    _enStatus = WaitZeroRPM;
  }
  else if( _enStatus == WaitZeroRPM &&  status == 7 /* stop */ && rpm == 0 )
  {
    _enStatus = Start; // once again
  }
}